use simcore_rs::{Time, SimContext, Facility, simulation, Process};
use rand::{distributions::Uniform, rngs::SmallRng, SeedableRng, Rng};
use criterion::{Criterion, BenchmarkId, criterion_group, criterion_main, BatchSize};

const SEED_A   :  u64 = 100000;
const SEED_S   :  u64 = 200000;
const RANGE    :  u32 =      8;
const STEP     : Time = 10000.0;

fn barbershop(c: &mut Criterion) {
	let mut group = c.benchmark_group("Barbershop");
	let mut rng_a = SmallRng::seed_from_u64(SEED_A);
	let mut rng_s = SmallRng::seed_from_u64(SEED_S);
	
	// group.measurement_time(Duration::from_secs_f64(60.0));
	group.confidence_level(0.99);
	
	for stop_time in (1..=RANGE).map(|c| Time::from(c)*STEP) {
		group.bench_with_input(
			BenchmarkId::from_parameter(stop_time),
			&stop_time,
			|b, &stop_time|
				b.iter_batched(
					|| BarberShop {
						stop_time,
						rng_a: SmallRng::from_rng(&mut rng_a).unwrap(),
						rng_s: SmallRng::from_rng(&mut rng_s).unwrap(),
						joe: Facility::new()
					},
					|shop| simulation(
						shop,
						|sim| Process::new(sim_main(sim))
					),
					BatchSize::SmallInput
				)
		);
	}
	
	group.finish();
}

criterion_group!(benches, barbershop);
criterion_main!(benches);

/* *************************** Barbershop Example *************************** */

/// Barbershop process.
struct BarberShop {
	stop_time: Time, rng_a: SmallRng, rng_s: SmallRng, joe: Facility
}

async fn sim_main(sim: SimContext<'_,BarberShop>) {
	// unpack the barber shop structure for easier access
	let stop_time = sim.shared().stop_time.clone();
	let mut rng_a = sim.shared().rng_a.clone();
	let mut rng_s = sim.shared().rng_s.clone();
	
	// activate a process to generate the customers
	sim.activate(async move {
		let dist = Uniform::new(12.0, 24.0);
		
		loop {
			// wait some time before activating the next customer
			sim.advance(rng_a.sample(dist)).await;
			
			// generate new customers until the store closes officially
			if sim.now() >= stop_time { return; }
			
			// activate the next customer
			sim.activate(Customer {
				rng: SmallRng::from_seed(rng_s.gen())
			}.actions(sim));
		}
	});
	
	// wait until the store closes
	sim.advance(stop_time).await;
	
	// finish processing the queue (no more customers arrive)
	sim.shared().joe.seize().await;
}

/// Customer process with access to the barber and a random number generator.
struct Customer { rng: SmallRng }

impl Customer {
	pub async fn actions(mut self, sim: SimContext<'_,BarberShop>) {
		// access the barber
		sim.shared().joe.seize().await;
		// spend time
		sim.advance(self.rng.gen_range(12.0, 18.0)).await;
		// release the barber
		sim.shared().joe.release();
	}
}
