use simcore_rs::{Process, Time, SimContext, Facility, RandomVar, simulation};
use rand::{distributions::Uniform, rngs::SmallRng, SeedableRng, Rng};

// helper constants
const STOP_TIME: Time = 480000.0;
const SEED_A   :  u64 = 100000;
const SEED_S   :  u64 = 200000;

/// Globally shared data.
struct Shared {
	joe: Facility,
	rv: RandomVar
}

/// Customer process with access to the barber and a random number generator.
struct Customer { rng: SmallRng }

impl Customer {
	pub async fn actions(mut self, sim: SimContext<'_,Shared>) {
		// access the barber and record the time for the report
		let time = sim.now();
		sim.shared().joe.seize().await;
		sim.shared().rv.tabulate(sim.now() - time);
		
		// spend time
		sim.advance(self.rng.gen_range(12.0, 18.0)).await;
		// release the barber
		sim.shared().joe.release();
	}
}

async fn sim_main(sim: SimContext<'_,Shared>) {
	// pseudo random number generators referenced from within the main process
	let mut rng_a = SmallRng::seed_from_u64(SEED_A);
	let mut rng_s = SmallRng::seed_from_u64(SEED_S);
	
	// activate a process to generate the customers
	sim.activate(async move {
		let dist = Uniform::new(12.0, 24.0);
		loop {
			sim.advance(rng_a.sample(dist)).await;
			if sim.now() >= STOP_TIME { return; }
			sim.activate(Customer {
				rng: SmallRng::from_seed(rng_s.gen())
			}.actions(sim));
		}
	});
	
	// wait until the store closes
	sim.advance(STOP_TIME).await;
	
	// finish processing the queue (no more customers arrive)
	sim.shared().joe.seize().await;
	
	// print statistics
	println!("Stats: {:#.3}", sim.shared().rv);
}

fn main() {
	simulation(
		// global data
		Shared { joe: Facility::new(), rv: RandomVar::new() }, 
		// simulation entry point
		|sim| Process::new(sim_main(sim))
	);
}
