///! Mini-library for the SAM 2020 paper.

use std::{
	collections::{BinaryHeap, VecDeque},
	task::{self, Context, Poll},
	fmt::{Display, Formatter},
	cell::{RefCell, Cell},
	future::Future,
	cmp::Ordering,
	pin::Pin,
	rc::{Rc, Weak}
};

// simple time type
pub type Time = f64;

// priority queue of time-process-pairs using time as the key
type EventQ<'p> = BinaryHeap<NextEvent<'p>>;

/// The (private) scheduler for processes.
struct Scheduler<'s,G> {
	/// The current simulation time.
	now: Cell<Time>,
	
	/// The event-calendar organized chronologically.
	calendar: RefCell<EventQ<'s>>,
	
	/// The currently active process.
	active: RefCell<Process<'s>>,
	
	/// Globally-accessible data.
	shared: G
}

// these allow us to *move* the context without invalidating it
/// A light-weight handle to the scheduler.
pub struct SimContext<'s,G> { handle: *const Scheduler<'s,G> }

impl<'s,G> Clone for SimContext<'s,G> {
	fn clone(&self) -> Self { *self }
}

impl<'s,G> Copy for SimContext<'s,G> {}

/// Performs a single simulation run.
/// 
/// Input is a function that takes a simulation context and returns the first
/// process. It would be better if the function only needed to return the future
/// needed to initialize the first process, but then the function signature gets
/// more complicated and is harder to explain in a paper.
pub fn simulation<G>(shared: G, main: impl FnOnce(SimContext<G>) -> Process) {
	// create a fresh scheduler and a handle to it
	let sched = Scheduler::new(shared);
	let sim = SimContext { handle: &sched };
	
	// construct a custom context to pass into the poll()-method
	let event_waker = StateEventWaker::new(sim);
	let waker = unsafe { event_waker.as_waker() };
	let mut cx = Context::from_waker(&waker);
	
	// evaluate the passed function for the main process and schedule it
	let root = main(sim);
	sched.schedule(root.clone());
	
	// pop processes until empty or the main process terminates
	while let Some(process) = sched.next_event() {
		if process.poll(&mut cx).is_ready() && process == root { break; }
	}
	
	// clear the scheduler before it is dropped to break the dependency
	// cycle between the processes and the scheduler that we made possible
	// by using a raw pointer in the simulation context
	sched.clear();
}

impl<'s,G> Scheduler<'s,G> {
	/// Creates a new scheduler.
	fn new(shared: G) -> Self {
		Self {
			now: Cell::default(),
			calendar: RefCell::default(),
			active: RefCell::default(),
			shared
		}
	}
	
	/// Clears the scheduler.
	fn clear(&self) {
		self.active.replace(Process::default());
		self.calendar.replace(EventQ::default());
	}
	
	/// Schedules a process at the current simulation time.
	fn schedule(&self, process: Process<'s>) {
		self.schedule_in(Time::default(), process);
	}
	
	/// Schedules a process at a later simulation time.
	fn schedule_in(&self, dt: Time, process: Process<'s>) {
		self.calendar.borrow_mut().push(
			NextEvent(self.now.get() + dt, process)
		);
	}
	
	/// Removes the process with the next event time from the calendar and
	/// activates it.
	fn next_event(&self) -> Option<Process<'s>> {
		let NextEvent(now, process) = self.calendar.borrow_mut().pop()?;
		self.now.set(now);
		self.active.replace(process.clone());
		Some(process)
	}
}

impl<'s,G> SimContext<'s,G> {
	/// Returns a (reference-counted) copy of the currently active process.
	pub fn active(&self) -> Process<'s> {
		self.sched().active.borrow().clone()
	}
	
	/// Activates a new process with the given future.
	pub fn activate(&self, fut: impl Future<Output = ()> + 's) {
		self.sched().schedule(Process::new(fut));
	}
	
	/// Reactivates a process that has been suspended with wait().
	pub fn reactivate(&self, process: Process<'s>) {
		self.sched().schedule(process);
	}
	
	/// Unconditionally suspends execution until reactivation.
	pub fn wait(&self) -> impl Future<Output = ()> {
		Wait { ready: false }
	}
	
	/// Reactivates the currently active process after some time has passed.
	pub async fn advance(&self, dt: Time) {
		self.sched().schedule_in(dt, self.active());
		self.wait().await
	}
	
	/// Returns the current simulation time.
	pub fn now(&self) -> Time {
		self.sched().now.get()
	}
	
	/// Returns a shared reference to the global data.
	pub fn shared(&self) -> &G {
		&self.sched().shared
	}
	
	/// Private function to get a safe reference to the scheduler.
	fn sched(&self) -> &Scheduler<'s,G> {
		// This is safe if no simulation context escapes from the closure
		// passed to simulation() which is enforced through the use of
		// higher-order trait-bounds.
		unsafe { &*self.handle }
	}
}

#[derive(Clone)]
/// Bare-bone process.
pub struct Process<'p>(Pin<Rc<RefCell<dyn Future<Output = ()> + 'p>>>);

impl<'p> Process<'p> {
	/// Creates a new process from a future.
	pub fn new(fut: impl Future<Output = ()> + 'p) -> Self {
		Process(Rc::pin(RefCell::new(fut)))
	}
	
	/// Private function for polling the future.
	fn poll(&self, cx: &mut Context) -> Poll<()> {
		// this is safe because the process is already pinned
		unsafe {
			Pin::new_unchecked(&mut *self.0.borrow_mut())
		}.poll(cx)
	}
}

// implementation of Rust's version of a default constructor
impl Default for Process<'_> {
	fn default() -> Self {
		Process::new(Wait { ready: false })
	}
}

// allows processes to be compared for equality
impl PartialEq for Process<'_> {
	fn eq(&self, other: &Self) -> bool {
		std::ptr::eq(&*self.0, &*other.0)
	}
}

// marks the equality-relation as total
impl Eq for Process<'_> {}

/// Time-process-pair that has a total order defined based on the time.
struct NextEvent<'p>(Time, Process<'p>);

impl PartialEq for NextEvent<'_> {
	fn eq(&self, other: &Self) -> bool {
		self.0 == other.0
	}
}

impl Eq for NextEvent<'_> {}

impl PartialOrd for NextEvent<'_> {
	fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
		Some(self.cmp(other))
	}
}

impl Ord for NextEvent<'_> {
	fn cmp(&self, other: &Self) -> Ordering {
		self.0.partial_cmp(&other.0)
			.expect("illegal event time NaN").reverse()
	}
}

/* **************************** specialized waker *************************** */

/// Complex waker that is used to implement state events.
/// 
/// This is the shallow version that is created on the stack of the function
/// running the event loop. It assumes that the stored process is the currently
/// active one and creates the deep version when it is cloned.
struct StateEventWaker<'s,G> { context: SimContext<'s,G> }

/// Deep version of the complex waker used to implement state events.
/// 
/// This version actually holds a process as well as a simulation context on the
/// heap. Dropping the waker releases the memory.
struct StateEventWakerDeep<'s,G> {
	process: Process<'s>,
	context: SimContext<'s,G>
}

impl<'s,G> StateEventWaker<'s,G> {
	/// Virtual function table for the waker.
	const VTABLE: task::RawWakerVTable = task::RawWakerVTable::new(
		Self::clone,
		Self::wake,
		Self::wake_by_ref,
		Self::drop
	);
	
	/// Creates a new (shallow) waker using only the simulation context.
	fn new(sim: SimContext<'s,G>) -> Self {
		StateEventWaker { context: sim }
	}
	
	/// Constructs a new waker using only a reference.
	/// 
	/// This function is unsafe because it is up to the user to ensure that the
	/// waker doesn't outlive the reference.
	unsafe fn as_waker(&self) -> task::Waker {
		task::Waker::from_raw(
			task::RawWaker::new(
				self as *const _ as *const (),
				&Self::VTABLE
			)
		)
	}
	
	/// Creates a deep waker from a reference.
	/// 
	/// This function is safe because it allocates memory on the heap.
	fn deep_waker(&self) -> task::RawWaker {
		StateEventWakerDeep::raw_waker(
			self.context, self.context.active()
		)
	}
	
	unsafe fn clone(this: *const ()) -> task::RawWaker {
		(&*(this as *const Self)).deep_waker()
	}
	
	unsafe fn wake(_this: *const ()) {
		unreachable!("attempting to awake the active process")
	}
	
	unsafe fn wake_by_ref(_this: *const ()) {
		unreachable!("attempting to awake the active process")
	}
	
	unsafe fn drop(_this: *const ()) {
		// memory released in the main event loop
	}
}

impl<'s,G> StateEventWakerDeep<'s,G> {
	/// Virtual function table for the deep waker.
	const VTABLE: task::RawWakerVTable = task::RawWakerVTable::new(
		Self::clone,
		Self::wake,
		Self::wake_by_ref,
		Self::drop
	);
	
	/// Constructs a raw waker from a simulation context and a process.
	fn raw_waker(sim: SimContext<'s,G>, process: Process<'s>) -> task::RawWaker {
		let waker = Box::leak(Box::new(
			Self { process, context: sim }
		));
		
		task::RawWaker::new(
			waker as *const _ as *const (),
			&Self::VTABLE
		)
	}
	
	unsafe fn clone(this: *const ()) -> task::RawWaker {
		let waker = &*(this as *const Self);
		Self::raw_waker(waker.context, waker.process.clone())
	}
	
	unsafe fn wake(this: *const ()) {
		let waker = Box::from_raw(this as *const Self as *mut Self);
		waker.context.reactivate(waker.process);
	}
	
	unsafe fn wake_by_ref(this: *const ()) {
		let waker = &*(this as *const Self);
		waker.context.reactivate(waker.process.clone());
	}
	
	unsafe fn drop(this: *const ()) {
		Box::from_raw(this as *const Self as *mut Self);
	}
}

/* *************************** specialized futures ************************** */

/// Future that blocks on the first call and returns on the second one.
struct Wait { ready: bool }

impl Future for Wait {
	type Output = ();
	
	fn poll(mut self: Pin<&mut Self>, _: &mut Context) -> Poll<Self::Output> {
		if self.ready {
			Poll::Ready(())
		} else {
			self.ready = true;
			Poll::Pending
		}
	}
}

/// Future that returns immediately with a cloned waker.
struct Waker;

impl Future for Waker {
	type Output = task::Waker;
	
	fn poll(self: Pin<&mut Self>, cx: &mut Context) -> Poll<Self::Output> {
		Poll::Ready(cx.waker().clone())
	}
}

// this is the end of the simulator core, the remainder is used in the example

/* *********************** synchronization structures *********************** */

/// GPSS-inspired facility that houses one exclusive resources to be used by
/// arbitrary many processes.
pub struct Facility {
	queue: RefCell<VecDeque<task::Waker>>,
	in_use: Cell<bool>
}

impl Facility {
	/// Creates a new facility.
	pub fn new() -> Self {
		Facility {
			queue: RefCell::new(VecDeque::new()),
			in_use: Cell::new(false)
		}
	}
	
	/// Resets the facility to its beginning state.
	pub fn reset(&self) {
		self.in_use.set(false);
		self.queue.borrow_mut().clear();
	}
	
	/// Attempts to seize the facility, blocking until it's possible.
	pub async fn seize(&self) {
		if self.in_use.replace(true) {
			self.queue.borrow_mut().push_back(Waker.await);
			Wait { ready: false }.await;
		}
	}
	
	/// Releases the facility and activates the next waiting process.
	pub fn release(&self) {
		if let Some(process) = self.queue.borrow_mut().pop_front() {
			process.wake();
		} else {
			self.in_use.set(false);
		}
	}
}

/// Complex channel with space for infinitely many elements of arbitrary
/// (non-process) type.
///
/// This channel supports arbitrary many readers and writers and uses wakers
/// to reactivate suspended processes.
struct Channel<T> {
	/// A queue of messages.
	store: VecDeque<T>,
	/// A queue of waiting pucks.
	waiting: VecDeque<task::Waker>
}

/// Creates a channel and returns a pair of read and write ends.
pub fn channel<T>() -> (Sender<T>, Receiver<T>) {
	let channel = Rc::new(RefCell::new(Channel {
		store: VecDeque::new(),
		waiting: VecDeque::new()
	}));
	(Sender(Rc::downgrade(&channel)), Receiver(channel))
}

/// Write-end of a channel.
pub struct Sender<T>(Weak<RefCell<Channel<T>>>);

/// Read-end of a channel.
pub struct Receiver<T>(Rc<RefCell<Channel<T>>>);

// Allow senders to be duplicated.
impl<T> Clone for Sender<T> {
	fn clone(&self) -> Self {
		Self(self.0.clone())
	}
}

// Allow receivers to be duplicated.
impl<T> Clone for Receiver<T> {
	fn clone(&self) -> Self {
		Self(self.0.clone())
	}
}

impl<T: Unpin> Sender<T> {
	/// Returns a future that can be awaited to send a message.
	pub fn send(&self, elem: T) -> SendFuture<T> {
		SendFuture(&self.0, Some(elem))
	}
}

impl<T: Unpin> Receiver<T> {
	/// Returns a future that can be awaited to receive a message.
	pub fn recv(&self) -> ReceiveFuture<T> {
		ReceiveFuture(&self.0)
	}
}

impl<T> Channel<T> {
	/// Private method enqueuing a process into the waiting list.
	fn enqueue(&mut self, process: task::Waker) {
		self.waiting.push_back(process);
	}
	
	/// Private method removing a process from the waiting list.
	fn dequeue(&mut self) -> Option<task::Waker> {
		self.waiting.pop_front()
	}
	
	/// Private method inserting a message into the queue.
	fn send(&mut self, value: T) {
		self.store.push_back(value);
	}
	
	/// Private method extracting a message from the queue non-blocking.
	fn recv(&mut self) -> Option<T> {
		self.store.pop_front()
	}
}

/// Future for the [`send()`] operation on a channel sender.
/// 
/// [`send()`]: struct.Sender.html#method.send
pub struct SendFuture<'c,T>(&'c Weak<RefCell<Channel<T>>>, Option<T>);

/// Future for the [`recv()`] operation on a channel receiver.
/// 
/// [`recv()`]: struct.Receiver.html#method.recv
pub struct ReceiveFuture<'c,T>(&'c Rc<RefCell<Channel<T>>>);

impl<T: Unpin> Future for SendFuture<'_,T> {
	type Output = Result<(),T>;
	
	fn poll(mut self: Pin<&mut Self>, _: &mut Context<'_>) -> Poll<Self::Output> {
		let elem = self.1.take().unwrap();
		
		// test if there are still readers left to listen
		if let Some(channel) = self.0.upgrade() {
			let mut channel = channel.borrow_mut();
			channel.send(elem);
			
			// awake a waiting process
			if let Some(process) = channel.dequeue() {
				process.wake();
			}
			
			Poll::Ready(Ok(()))
		} else {
			Poll::Ready(Err(elem))
		}
	}
}

impl<T: Unpin> Future for ReceiveFuture<'_,T> {
	type Output = Option<T>;
	
	fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
		let mut channel = self.0.borrow_mut();
		
		if let Some(c) = channel.recv() {
			// there are elements in the channel
			Poll::Ready(Some(c))
		} else if Rc::weak_count(&self.0) == 0 {
			// no elements in the channel and no potential senders exist
			Poll::Ready(None)
		} else {
			// no elements, but potential senders exist: check back later
			channel.enqueue(cx.waker().clone());
			Poll::Pending
		}
	}
}

/* ************************* statistical facilities ************************* */

/// A simple collector for statistical data, inspired by SLX's random_variable.
pub struct RandomVar {
	total: Cell<u32>, sum: Cell<f64>, sqr: Cell<f64>,
	min: Cell<f64>, max: Cell<f64>
}

impl RandomVar {
	/// Creates a new random variable.
	pub fn new() -> Self {
		RandomVar::default()
	}
	
	/// Resets all stored statistical data.
	pub fn clear(&self) {
		self.total.set(0);
		self.sum.set(0.0);
		self.sqr.set(0.0);
		self.min.set(f64::INFINITY);
		self.max.set(f64::NEG_INFINITY);
	}
	
	/// Adds another value to the statistical collection.
	pub fn tabulate<T: Into<f64>>(&self, val: T) {
		let val: f64 = val.into();
		
		self.total.set(self.total.get() + 1);
		self.sum.set(self.sum.get() + val);
		self.sqr.set(self.sqr.get() + val*val);
		
		if self.min.get() > val {
			self.min.set(val);
		} else if self.max.get() < val {
			self.max.set(val);
		}
	}
}

impl Default for RandomVar {
	fn default() -> Self {
		RandomVar {
			total: Cell::default(),
			sum: Cell::default(),
			sqr: Cell::default(),
			min: Cell::new(f64::INFINITY),
			max: Cell::new(f64::NEG_INFINITY)
		}
	}
}

impl Display for RandomVar {
	fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
		let total = self.total.get();
		let mean = self.sum.get() / f64::from(total);
		let variance = self.sqr.get() / f64::from(total) - mean*mean;
		let std_dev = variance.sqrt();
		
		f.debug_struct("RandomVar")
			.field("total", &total)
			.field("mean", &mean)
			.field("std_dev", &std_dev)
			.field("min", &self.min.get())
			.field("max", &self.max.get())
			.finish()
	}
}

